from distutils.core import setup, Extension

include_dirs = []
library_dirs = []
libraries = []
runtime_library_dirs = []
extra_objects = []
define_macros = []

setup(name = "pyqdbm",
      version = "0.90",
      author = "Yoshitaka Hirano",
      author_email = "yo@hirano.cc",
      license = "shiran",
      url = "http://www.hirano.cc/soft/pyqdbm.html",
      packages = ["qdbm"],
      ext_package = "qdbm",
      ext_modules = [Extension( name = "depot",
                                sources = ["src/depot.c"],
                                include_dirs = include_dirs,
                                library_dirs = library_dirs,
                                runtime_library_dirs = runtime_library_dirs,
                                libraries = libraries,
                                extra_objects = extra_objects,
                                define_macros = define_macros
                              )],
      )

